@@ -26,9 +26,11 @@
 #include "blocks.h"
 #include "preview.h"
 #include "scoreframe.h"
+#include "sound.h"
 
 #include <games-gconf.h>
 #include <games-frame.h>
+#include <games-controls.h>
 
 #include <libgnomevfs/gnome-vfs.h>
 
@@ -62,6 +64,7 @@ bool rotateCounterClockWise = true;
 #define TETRIS_OBJECT "gnometris-tetris-object"
 
 #define KEY_OPTIONS_DIR "/apps/gnometris/options"
+#define KEY_SOUND "/apps/gnometris/options/sound"
 #define KEY_BLOCK_PIXMAP "/apps/gnometris/options/block_pixmap"
 #define KEY_STARTING_LEVEL "/apps/gnometris/options/starting_level"
 #define KEY_DO_PREVIEW "/apps/gnometris/options/do_preview"
@@ -70,8 +73,26 @@ bool rotateCounterClockWise = true;
 #define KEY_LINE_FILL_HEIGHT "/apps/gnometris/options/line_fill_height"
 #define KEY_LINE_FILL_PROBABILITY "/apps/gnometris/options/line_fill_probability"
 
-#define URI_LIST 0
-#define TEXT_PLAIN 1
+#define KEY_CONTROLS_DIR "/apps/gnometris/controls"
+#define KEY_MOVE_LEFT "/apps/gnometris/controls/key_left"
+#define KEY_MOVE_RIGHT "/apps/gnometris/controls/key_right"
+#define KEY_MOVE_DOWN "/apps/gnometris/controls/key_down"
+#define KEY_MOVE_DROP "/apps/gnometris/controls/key_drop"
+#define KEY_MOVE_ROTATE "/apps/gnometris/controls/key_rotate"
+#define KEY_MOVE_PAUSE "/apps/gnometris/controls/key_pause"
+
+#define KEY_BG_COLOUR "/apps/gnometris/options/bgcolor"
+#define KEY_USE_BG_IMAGE "/apps/gnometris/options/usebgimage"
+
+#define TILE_THRESHOLD 65
+
+enum {
+	URI_LIST,
+	TEXT_PLAIN,
+	COLOUR,
+	RESET
+};
+
 
 Tetris::Tetris(int cmdlLevel): 
 	blockPixmap(0),
@@ -93,8 +114,14 @@ Tetris::Tetris(int cmdlLevel):
 	double pts;
 	gchar * outdir;
 	GtkTargetEntry targets[] = {{"text/uri-list", 0, URI_LIST}, 
+				    {"property/bgimage", 0, URI_LIST},
 				    {"text/plain", 0, TEXT_PLAIN},
-				    {"STRING", 0, TEXT_PLAIN}};
+				    {"STRING", 0, TEXT_PLAIN},
+				    {"application/x-color", 0, COLOUR},
+				    {"x-special/gnome-reset-background", 0, RESET}};
+
+	if (!sound)
+		sound = new Sound();
 
 	pic = new GdkPixbuf*[tableSize];
 	for (int i = 0; i < tableSize; ++i)
@@ -163,6 +190,9 @@ Tetris::Tetris(int cmdlLevel):
 	gconf_client_add_dir (gconf_client, KEY_OPTIONS_DIR, GCONF_CLIENT_PRELOAD_NONE, NULL);
 	gconf_client_notify_add (gconf_client, KEY_OPTIONS_DIR, gconfNotify, this, NULL, NULL);
 
+	gconf_client_add_dir (gconf_client, KEY_CONTROLS_DIR, GCONF_CLIENT_PRELOAD_NONE, NULL);
+	gconf_client_notify_add (gconf_client, KEY_CONTROLS_DIR, gconfNotify, this, NULL, NULL);
+
 	initOptions ();
 
 	gnome_app_create_menus(GNOME_APP(w), mainmenu);
@@ -183,7 +213,8 @@ Tetris::Tetris(int cmdlLevel):
 	field->show();
 	setupPixmap();
 
-	g_signal_connect (w, "event", G_CALLBACK (eventHandler), this);
+	g_signal_connect (w, "key_press_event", G_CALLBACK (keyPressHandler), this);
+	g_signal_connect (w, "key_release_event", G_CALLBACK (keyReleaseHandler), this);
   
 	GtkWidget *vb2 = gtk_vbox_new(FALSE, 0);
 	gtk_container_set_border_width(GTK_CONTAINER(vb2), 10);
@@ -259,7 +290,6 @@ Tetris::Tetris(int cmdlLevel):
 	gnome_canvas_item_set (gameoverMessage, "size_points", pts, 0);
 
         gnome_canvas_item_hide (gameoverMessage);
-
 }
 
 Tetris::~Tetris()
@@ -328,12 +358,12 @@ Tetris::setupPixmap()
 
 	if (!g_file_test (fullpixname, G_FILE_TEST_EXISTS))
 	{
-		GtkWidget *w = gtk_message_dialog_new (NULL,
+		GtkWidget *widget = gtk_message_dialog_new (NULL,
 						       GTK_DIALOG_DESTROY_WITH_PARENT,
 						       GTK_MESSAGE_ERROR,
 						       GTK_BUTTONS_OK,
 						       _("Could not find the theme: \n%s\n\nPlease check your gnome-games installation"), fullpixname);
-		gtk_dialog_run (GTK_DIALOG (w));
+		gtk_dialog_run (GTK_DIALOG (widget));
 		exit(1);
 	}
 
@@ -343,12 +373,12 @@ Tetris::setupPixmap()
 	image = gdk_pixbuf_new_from_file(fullpixname, NULL);
 
 	if (image == NULL) {
-		GtkWidget *w = gtk_message_dialog_new (NULL,
+		GtkWidget *widget= gtk_message_dialog_new (NULL,
 						       GTK_DIALOG_DESTROY_WITH_PARENT,
 						       GTK_MESSAGE_ERROR,
 						       GTK_BUTTONS_OK,
 						       _("Can't load the image: \n%s\n\nPlease check your gnome-games installation"), fullpixname);
-		gtk_dialog_run (GTK_DIALOG (w));
+		gtk_dialog_run (GTK_DIALOG (widget));
 		exit (1);
 	}
 	g_free (fullpixname);
@@ -371,14 +401,61 @@ Tetris::setupPixmap()
 	if (bgimage)
 		g_object_unref (G_OBJECT (bgimage));
 
-	if (g_file_test (bgPixmap, G_FILE_TEST_EXISTS)) 
-		bgimage = gdk_pixbuf_new_from_file (bgPixmap, NULL);
-	else 
+	if (!usebg)
 		bgimage = NULL;
+	else {
+		if (g_file_test (bgPixmap, G_FILE_TEST_EXISTS)) 
+			bgimage = gdk_pixbuf_new_from_file (bgPixmap, NULL);
+		else 
+			bgimage = NULL;
+	}
+
+	/* A nasty hack to tile the image if it looks tileable (i.e. it
+	 * is small enough. */
+	if (bgimage) {
+		int width, height;
+		int bgwidth, bgheight;
+			
+		bgwidth = COLUMNS*BLOCK_SIZE;
+		bgheight = LINES*BLOCK_SIZE;
+
+		width = gdk_pixbuf_get_width (bgimage);
+		height = gdk_pixbuf_get_height (bgimage);
+
+		/* The heuristic is, anything less than 65 pixels on a side,
+		 * or is square and smaller than the playing field is tiled. */
+		/* Note that this heuristic fails for the standard nautilus
+		 * background burlap.jpg because it is 97x91 */
+		if ((width < TILE_THRESHOLD) || (height < TILE_THRESHOLD) ||
+		    ((width == height) && (width < bgwidth))) {
+			GdkPixbuf * temp;
+			int i, j;
+			
+			temp = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, 
+					       bgwidth, bgheight); 
+		
+			for (i=0; i<=bgwidth/width; i++) {
+				for (j=0; j<=bgheight/height; j++) {
+					int x, y, ww, hh;
+
+					x = i*width;
+					y = j*height;
+					ww = MIN (width, bgwidth - x);
+					hh = MIN (height, bgheight - y);
+
+					gdk_pixbuf_copy_area (bgimage, 0, 0,
+							      ww, hh, temp,
+							      x, y);
+				}
+			}
+			g_object_unref (bgimage);
+			bgimage = temp;
+		}
+	}
 
 	if (field)
 	{
-		field->updateSize (bgimage);
+		field->updateSize (bgimage, &bgcolour);
 		gtk_widget_queue_resize (field->getWidget());
 	}
 	
@@ -398,76 +475,108 @@ Tetris::gconfNotify (GConfClient *tmp_client, guint cnx_id, GConfEntry *tmp_entr
 	t->setOptions ();
 }
 
+char *
+Tetris::gconfGetString (GConfClient *client, const char *key, const char *default_val)
+{
+	char *val;
+	GError *error = NULL;
+
+	val = gconf_client_get_string (client, key, &error);
+	if (error) {
+		g_warning ("gconf error: %s\n", error->message);
+		g_error_free (error);
+		error = NULL;
+		val = g_strdup (default_val);
+	}
+
+	return val;
+}
+
+int
+Tetris::gconfGetInt (GConfClient *client, const char *key, int default_val)
+{
+	int val;
+	GError *error = NULL;
+
+	val = gconf_client_get_int (client, key, &error);
+	if (error) {
+		g_warning ("gconf error: %s\n", error->message);
+		g_error_free (error);
+		error = NULL;
+		val = default_val;
+	}
+
+	return val;
+}
+
+gboolean
+Tetris::gconfGetBoolean (GConfClient *client, const char *key, gboolean default_val)
+{
+	gboolean val;
+	GError *error = NULL;
+
+	val = gconf_client_get_bool (client, key, &error);
+	if (error) {
+		g_warning ("gconf error: %s\n", error->message);
+		g_error_free (error);
+		error = NULL;
+		val = default_val;
+	}
+
+	return val;
+}
+
 void
 Tetris::initOptions ()
 {
-	GError *error = NULL;
+	gchar *bgcolourstr;
 
 	if (blockPixmap)
 		g_free (blockPixmap);
 
-	blockPixmap = gconf_client_get_string (gconf_client, KEY_BLOCK_PIXMAP, &error);
-	if (error) {
-		g_warning ("gconf error: %s\n", error->message);
-		g_error_free (error);
-		error = NULL;
-	}
-        if (blockPixmap == NULL)
-          blockPixmap = g_strdup ("7blocks-tig.png");
+	blockPixmap = gconfGetString (gconf_client, KEY_BLOCK_PIXMAP, "7blocks-tig.png");
 
-	startingLevel = gconf_client_get_int (gconf_client, KEY_STARTING_LEVEL, &error);
-	if (error) {
-		g_warning ("gconf error: %s\n", error->message);
-		g_error_free (error);
-		error = NULL;
-	}
-        if (startingLevel < 1) 
-          startingLevel = 1;
-        if (startingLevel > 10) 
-          startingLevel = 10;
+	startingLevel = gconfGetInt (gconf_client, KEY_STARTING_LEVEL, 1);
+	if (startingLevel < 1) 
+		startingLevel = 1;
+	if (startingLevel > 10) 
+		startingLevel = 10;
 
-	do_preview = gconf_client_get_bool (gconf_client, KEY_DO_PREVIEW, &error);
-	if (error) {
-		g_warning ("gconf error: %s\n", error->message);
-		g_error_free (error);
-		error = NULL;
-	}
+	if (gconfGetBoolean (gconf_client, KEY_SOUND, TRUE)) 
+		sound->turnOn ();
+	else
+		sound->turnOff ();
 
-	random_block_colors = gconf_client_get_bool (gconf_client, KEY_RANDOM_BLOCK_COLORS, &error);
-	if (error) {
-		g_warning ("gconf error: %s\n", error->message);
-		g_error_free (error);
-		error = NULL;
-	}
-	
-	rotateCounterClockWise = gconf_client_get_bool (gconf_client, KEY_ROTATE_COUNTER_CLOCKWISE, &error);
-	if (error) {
-		g_warning ("gconf error: %s\n", error->message);
-		g_error_free (error);
-		error = NULL;
-	}
+	do_preview = gconfGetBoolean (gconf_client, KEY_DO_PREVIEW, TRUE);
 
-	line_fill_height = gconf_client_get_int (gconf_client, KEY_LINE_FILL_HEIGHT, &error);
-	if (error) {
-		g_warning ("gconf error: %s\n", error->message);
-		g_error_free (error);
-		error = NULL;
-	}
+	random_block_colors = gconfGetBoolean (gconf_client, KEY_RANDOM_BLOCK_COLORS, TRUE);
+
+	rotateCounterClockWise = gconfGetBoolean (gconf_client, KEY_ROTATE_COUNTER_CLOCKWISE, TRUE);
+
+	line_fill_height = gconfGetInt (gconf_client, KEY_LINE_FILL_HEIGHT, 0);
 	if (line_fill_height < 0)
 		line_fill_height = 0;
 	if (line_fill_height > 19)
 		line_fill_height = 19;
-	
-	line_fill_prob = gconf_client_get_int (gconf_client, KEY_LINE_FILL_PROBABILITY, &error);
-	if (error) {
-		g_warning ("gconf error: %s\n", error->message);
-		g_error_free (error);
-		error = NULL;
-	}
+
+	line_fill_prob = gconfGetInt (gconf_client, KEY_LINE_FILL_PROBABILITY, 0);
 	if (line_fill_prob < 0)
 		line_fill_prob = 0;
 	if (line_fill_prob > 10)
 		line_fill_prob = 10;
+
+	moveLeft = gconfGetInt (gconf_client, KEY_MOVE_LEFT, GDK_Left);
+	moveRight = gconfGetInt (gconf_client, KEY_MOVE_RIGHT, GDK_Right);
+	moveDown = gconfGetInt (gconf_client, KEY_MOVE_DOWN, GDK_Down);
+	moveDrop = gconfGetInt (gconf_client, KEY_MOVE_DROP, GDK_Pause);
+	moveRotate = gconfGetInt (gconf_client, KEY_MOVE_ROTATE, GDK_Up);
+	movePause = gconfGetInt (gconf_client, KEY_MOVE_PAUSE, GDK_space);
+
+	bgcolourstr = gconfGetString (gconf_client, KEY_BG_COLOUR, "Black");
+	gdk_color_parse (bgcolourstr, &bgcolour);
+	g_free (bgcolourstr);
+
+	usebg = gconfGetBoolean (gconf_client, KEY_USE_BG_IMAGE, FALSE);
 }
 
 void
@@ -477,6 +586,7 @@ Tetris::setOptions ()
 		gtk_spin_button_set_value (GTK_SPIN_BUTTON (sentry), startingLevel);
 		gtk_spin_button_set_value (GTK_SPIN_BUTTON (fill_prob_spinner), line_fill_prob);
 		gtk_spin_button_set_value (GTK_SPIN_BUTTON (fill_height_spinner), line_fill_height);
+		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (sound_toggle), sound->isOn ());
 		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (do_preview_toggle), do_preview);
 		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (random_block_colors_toggle), random_block_colors);
 		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (rotate_counter_clock_wise_toggle), rotateCounterClockWise);
@@ -487,6 +597,14 @@ Tetris::setOptions ()
 	setupPixmap ();
 }
 
+void
+Tetris::setSound (GtkWidget *widget, gpointer data)
+{
+	Tetris *t = (Tetris*)data;
+	gconf_client_set_bool (t->gconf_client, KEY_SOUND,
+			       GTK_TOGGLE_BUTTON (widget)->active, NULL);
+}
+
 void 
 Tetris::setSelectionPreview(GtkWidget *widget, void *d)
 {
@@ -611,11 +729,14 @@ Tetris::fillMenu(GtkWidget *menu, char *pixname, char *dirname,
 int 
 Tetris::gameProperties(GtkWidget *widget, void *d)
 {
+	GtkWidget *notebook;
+	GtkWidget *vbox;
 	GtkWidget *label;
 	GtkWidget *frame;
 	GtkWidget *table;
 	GtkWidget *fvbox;
 	GtkObject *adj;
+	GtkWidget *controls_list;
         
 	Tetris *t = (Tetris*) d;
 	
@@ -638,6 +759,15 @@ Tetris::gameProperties(GtkWidget *widget, void *d)
 	g_signal_connect (t->setupdialog, "response",
 			  G_CALLBACK (setupdialogResponse), d);
 
+	notebook = gtk_notebook_new ();
+	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(t->setupdialog)->vbox), notebook,
+			    TRUE, TRUE, 0);
+
+	/* game page */
+	vbox = gtk_vbox_new (FALSE, 0);
+	label = gtk_label_new_with_mnemonic (_("_Game"));
+	gtk_notebook_append_page (GTK_NOTEBOOK(notebook), vbox, label);
+
 	frame = games_frame_new (_("Setup"));
 	table = gtk_table_new (3, 2, FALSE);
 	gtk_container_set_border_width (GTK_CONTAINER (table), 0);
@@ -689,12 +819,21 @@ Tetris::gameProperties(GtkWidget *widget, void *d)
 	gtk_table_attach_defaults (GTK_TABLE (table), t->sentry, 1, 2, 2, 3);
 
 	gtk_container_add (GTK_CONTAINER (frame), table);
-	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (t->setupdialog)->vbox), frame, 
+	gtk_box_pack_start (GTK_BOX (vbox), frame, 
 			    FALSE, FALSE, 0);
 
 	frame = games_frame_new (_("Operation"));
-	fvbox = gtk_vbox_new (FALSE, FALSE);
-	gtk_box_set_spacing (GTK_BOX (fvbox), 6);
+	fvbox = gtk_vbox_new (FALSE, 6);
+
+	/* sound */
+	t->sound_toggle =
+		gtk_check_button_new_with_label (_("Enable sounds"));
+	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON (t->sound_toggle),
+				     sound->isOn ());
+	g_signal_connect (t->sound_toggle, "clicked",
+			  G_CALLBACK (setSound), d);
+	gtk_box_pack_start (GTK_BOX (fvbox), t->sound_toggle, 0, 0, 0);
+	
 
 	/* preview next block */
 	t->do_preview_toggle =
@@ -728,7 +867,7 @@ Tetris::gameProperties(GtkWidget *widget, void *d)
 			    0, 0, 0);
 
 	gtk_container_add (GTK_CONTAINER (frame), fvbox);
-	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (t->setupdialog)->vbox), frame, 
+	gtk_box_pack_start (GTK_BOX (vbox), frame, 
 			    FALSE, FALSE, 0);
 
 	frame = games_frame_new (_("Theme"));
@@ -749,8 +888,31 @@ Tetris::gameProperties(GtkWidget *widget, void *d)
 	gtk_table_attach_defaults (GTK_TABLE (table), omenu, 1, 2, 0, 1);
 
 	gtk_container_add (GTK_CONTAINER (frame), table);
-	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (t->setupdialog)->vbox), frame, 
+	gtk_box_pack_start (GTK_BOX (vbox), frame, 
 			    FALSE, FALSE, 0);
+
+	/* controls page */
+	vbox = gtk_vbox_new (FALSE, 0);
+	label = gtk_label_new_with_mnemonic (_("_Controls"));
+	gtk_notebook_append_page (GTK_NOTEBOOK(notebook), vbox, label);
+
+	frame = games_frame_new (_("Keyboard controls"));
+	fvbox = gtk_vbox_new (FALSE, 6);
+	gtk_container_add (GTK_CONTAINER (frame), fvbox);
+
+	controls_list = games_controls_list_new ();
+	games_controls_list_add_controls (GAMES_CONTROLS_LIST (controls_list),
+					  KEY_MOVE_LEFT,
+					  KEY_MOVE_RIGHT,
+					  KEY_MOVE_DOWN,
+					  KEY_MOVE_DROP,
+					  KEY_MOVE_ROTATE,
+					  KEY_MOVE_PAUSE,
+					  NULL);
+
+	gtk_box_pack_start (GTK_BOX (fvbox), controls_list, TRUE, TRUE, 0);
+
+	gtk_container_add (GTK_CONTAINER (vbox), frame);
 	
 	gtk_widget_show_all (t->setupdialog);
 	gtk_widget_set_sensitive(t->gameMenuPtr[0].widget, FALSE);
@@ -769,7 +931,7 @@ int
 Tetris::gameEnd(GtkWidget *widget, void *d)
 {
 	Tetris *t = (Tetris*) d;
-	
+
 	g_source_remove(t->timeoutId);
 	t->timeoutId = 0;
 	blocknr_next = -1;
@@ -809,10 +971,13 @@ void
 Tetris::manageFallen()
 {
 	ops->fallingToLaying();
+	sound->playSound (SOUND_LAND);
 
 	int levelBefore = scoreFrame->getLevel();
 	ops->checkFullLines(scoreFrame);
 	int levelAfter = scoreFrame->getLevel();
+	if (levelAfter != levelBefore) 
+		sound->playSound (SOUND_GNOMETRIS);
 	if ((levelBefore != levelAfter) || fastFall)
 		generateTimer(levelAfter);
 	
@@ -852,108 +1017,123 @@ Tetris::timeoutHandler(void *d)
 	return TRUE;	
 }
 
-gint
-Tetris::eventHandler(GtkWidget *widget, GdkEvent *event, void *d)
+gboolean
+Tetris::keyPressHandler(GtkWidget *widget, GdkEvent *event, Tetris *t)
 {
-	Tetris *t = (Tetris*) d;
-        int bonus;
-        
+	int keyval;
+	bool res = false;
+	int bonus;
+
 	if (t->timeoutId == 0)
 		return FALSE;
-	
-	if (event->type == GDK_KEY_PRESS)
-	{
-		int keyval = ((GdkEventKey*)event)->keyval;
 
-		if ((keyval == GDK_p) || (keyval == GDK_P))
-		{
-			t->togglePause();
-			return TRUE;
-		}
+	keyval = ((GdkEventKey*)event)->keyval;
+
+	if (keyval == t->movePause)
+	{
+		t->togglePause();
+		return TRUE;
 	}
+
+	if (t->paused)
+		return FALSE;
+
+	if (keyval == t->moveLeft) {
+		sound->playSound (SOUND_SLIDE);
+		res = t->ops->moveBlockLeft();
+		t->onePause = false;
+	} else if (keyval == t->moveRight) {
+		sound->playSound (SOUND_SLIDE);
+		res = t->ops->moveBlockRight();
+		t->onePause = false;
+	} else if (keyval == t->moveRotate) {
+		sound->playSound (SOUND_TURN);
+		res = t->ops->rotateBlock(rotateCounterClockWise);
+		t->onePause = false;
+	} else if (keyval == t->moveDown) {
+		if (!t->fastFall && !t->onePause) {
+			t->fastFall = true;
+			t->fastFallPoints = 0;
+			g_source_remove (t->timeoutId);
+			t->timeoutId = g_timeout_add_full(0, 10, timeoutHandler, t, NULL);
+			res = true;
+		}
+	} else if (keyval == t->moveDrop) {
+		if (!t->dropBlock) {
+			t->dropBlock = true;
+			bonus = t->ops->dropBlock();
+			t->scoreFrame->incScore(bonus);
+			t->manageFallen();
+			res = TRUE;
+		}        
+	}
+
+	return res;
+}
+
+gint
+Tetris::keyReleaseHandler(GtkWidget *widget, GdkEvent *event, Tetris *t)
+{
+	bool res = false;
+
+	if (t->timeoutId == 0)
+		return FALSE;
 	
 	if (t->paused)
 		return FALSE;
 
-	bool res = false;
-	bool keyEvent = false;
+	int keyval = ((GdkEventKey*)event)->keyval;
 
-	switch (event->type)
-	{
-	case GDK_KEY_PRESS: 
-	{
-		GdkEventKey *e = (GdkEventKey*)event;
-		keyEvent = true;
-		
-		switch(e->keyval)
-		{
-		case GDK_Left:
-			res = t->ops->moveBlockLeft();
-                        t->onePause = false;
-			break;
-		case GDK_Right:
-			res = t->ops->moveBlockRight();
-                        t->onePause = false;
-			break;
-		case GDK_Up:
-			res = t->ops->rotateBlock(rotateCounterClockWise);
-                        t->onePause = false;
-			break;
-		case GDK_Down:
-			if (!t->fastFall && !t->onePause)
-			{
-				t->fastFall = true;
-				t->fastFallPoints = 0;
-				g_source_remove(t->timeoutId);
-				t->timeoutId = g_timeout_add_full(0, 10, timeoutHandler, t, NULL);
-				res = true;
-			}
-			break;
-		case GDK_space:
-                        if (!t->dropBlock)
-                        {
-                                t->dropBlock = true;
-                                bonus = t->ops->dropBlock();
-                                t->scoreFrame->incScore(bonus);
-                                t->manageFallen();
-                                res = TRUE;
-                        }        
-			break;
-		default:
-			return FALSE;
+	if (keyval == t->moveDown) {
+		if (t->fastFall) {
+			t->fastFall = false;
+ 			t->generateTimer(t->scoreFrame->getLevel());
 		}
-		break;
-	}
-	case GDK_KEY_RELEASE:
-	{
-		GdkEventKey *e = (GdkEventKey*)event;
-
-		switch (e->keyval)
-		{
-		case GDK_Down:
-			keyEvent = true;
-			if (t->fastFall)
-			{
-				t->fastFall = false;
- 				t->generateTimer(t->scoreFrame->getLevel());
-			}
-			break;
-		case GDK_space:
-                        t->dropBlock = false;
-			break;
-		default:
-			return FALSE;
-		}
-		break;
-	}
-	default:
-		break;
+		res = TRUE;
+	} else if (keyval == t->moveDrop) {
+		t->dropBlock = false;
+		res = TRUE;
 	}
 
-	if (res)
-		gtk_widget_queue_draw(t->field->getWidget());
+	return res;
+}
 
-	return (keyEvent == true);
+void Tetris::saveBgOptions ()
+{
+	gchar * cbuffer;
+
+	gconf_client_set_bool (gconf_client, KEY_USE_BG_IMAGE, 
+				  usebg, NULL);
+
+	cbuffer = g_strdup_printf ("#%04x%04x%04x", bgcolour.red,
+				   bgcolour.green, bgcolour.blue);
+	gconf_client_set_string (gconf_client, KEY_BG_COLOUR, cbuffer,
+				 NULL);
+	g_free (cbuffer);
+}
+
+void
+Tetris::decodeColour (guint16 *data, Tetris *t)
+{
+	t->bgcolour.red = data[0];
+	t->bgcolour.green = data[1];
+	t->bgcolour.blue = data[2];
+	/* Ignore the alpha channel. */
+
+	t->usebg = FALSE;
+	t->saveBgOptions ();
+}
+
+void
+Tetris::resetColour (Tetris *t)
+{
+	t->bgcolour.red = 0;
+	t->bgcolour.green = 0;
+	t->bgcolour.blue = 0;
+	/* Ignore the alpha channel. */
+
+	t->usebg = FALSE;
+	t->saveBgOptions ();
 }
 
 gchar * 
@@ -983,7 +1163,7 @@ Tetris::decodeDropData(gchar * data, gint type)
 
 		/* Now extract the first URI. */
 		end = start;
-		while ((*end != '\0') && (*end != '\r'))
+		while ((*end != '\0') && (*end != '\r') && (*end != '\n'))
 			end++;
 		*end = '\0';
 
@@ -1009,16 +1189,18 @@ Tetris::dragDrop(GtkWidget *widget, GdkDragContext *context,
 	GdkPixbuf * pixbuf;
 	guchar * buffer;
 
+
 	/* Accept a dropped filename and try and load it as the
 	   background image. In the event of any kind of failure we
 	   silently ignore it. */
 	
-	/* FIXME: We should also handle dropped colours so we get a
-	   solid background (dropped gimp gradients too ?). */
+	/* FIXME: We don't handle colour gradients (e.g. from the gimp) */
 
-	/* FIXME: We don't accept nautlus bgs/colours either. */
+	/* FIXME: Drag and drop from konqueror doesn't work, we
+	 * aren't even registering that it is providing test/uri-list
+	 * content. */
 
-	/* Drag and drop from eog isn't working either. */
+	/* FIXME: Dropped URLs from mozilla don't work (see below). */
 
 	if (data->length < 0) {
 		gtk_drag_finish (context, FALSE, FALSE, time);
@@ -1027,6 +1209,17 @@ Tetris::dragDrop(GtkWidget *widget, GdkDragContext *context,
 
 	gtk_drag_finish (context, TRUE, FALSE, time);
 
+	if (info == COLOUR) {
+		if (data->length == 8) 
+			decodeColour ((guint16 *)data->data, t);
+		return;
+	}
+
+	if (info == RESET) {
+		resetColour (t);
+		return;
+	}
+
 	fileuri = decodeDropData ((gchar *)data->data, info);
 	/* Silently ignore bad data. */
 	if (fileuri == NULL)
@@ -1047,6 +1240,8 @@ Tetris::dragDrop(GtkWidget *widget, GdkDragContext *context,
 	if (result != GNOME_VFS_OK)
 		goto error_exit_handle;
 
+	/* This is where Drag and Drop of URLs from mozilla fails. */
+
 	if (!(fileinfo.valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SIZE))
 		goto error_exit_handle;
 
@@ -1087,7 +1282,8 @@ Tetris::dragDrop(GtkWidget *widget, GdkDragContext *context,
 	if ((result != GNOME_VFS_OK) || (bytesread != filesize))
 	    goto error_exit_saver;
 
-	t->setupPixmap ();
+	t->usebg = TRUE;
+	t->saveBgOptions ();
 
  error_exit_saver:
 	gnome_vfs_close (outhandle);
@@ -1144,6 +1340,7 @@ Tetris::endOfGame()
         gnome_canvas_item_hide (pauseMessage);
         gnome_canvas_item_show (gameoverMessage);
         gnome_canvas_item_raise_to_top (gameoverMessage);
+	sound->playSound (SOUND_GAMEOVER);
 	inPlay = false;
 
 	if (scoreFrame->getScore() > 0) 
@@ -1207,6 +1404,8 @@ Tetris::gameNew(GtkWidget *widget, void *d)
         gnome_canvas_item_hide (t->pauseMessage);
         gnome_canvas_item_hide (t->gameoverMessage);
         
+	sound->playSound (SOUND_GNOMETRIS);
+
 	return TRUE;
 }
 
